"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getZetaPerpMarkets = exports.executeAndRetry = exports.cleanupChannel = exports.zetaMarketsChannel = exports.zetaDataChannel = exports.zetaProducerReadyChannel = exports.minionReadyChannel = exports.CircularBuffer = exports.decimalPlaces = exports.batch = exports.getAllowedValuesText = exports.getDidYouMean = exports.wait = void 0;
const web3_js_1 = require("@solana/web3.js");
const sdk_1 = require("@zetamarkets/sdk");
const assets_1 = require("@zetamarkets/sdk/dist/assets");
const didyoumean2_1 = __importDefault(require("didyoumean2"));
const wait = (delayMS) => new Promise((resolve) => setTimeout(resolve, delayMS));
exports.wait = wait;
function getDidYouMean(input, allowedValues) {
    let tip = '';
    if (typeof input === 'string') {
        let result = (0, didyoumean2_1.default)(input, allowedValues, {});
        if (result !== null) {
            tip = ` Did you mean '${result}'?`;
        }
    }
    return tip;
}
exports.getDidYouMean = getDidYouMean;
function getAllowedValuesText(allowedValues) {
    return `Allowed values: ${allowedValues.map((val) => `'${val}'`).join(', ')}.`;
}
exports.getAllowedValuesText = getAllowedValuesText;
function* batch(items, batchSize) {
    for (let i = 0; i < items.length; i += batchSize) {
        yield items.slice(i, i + batchSize);
    }
}
exports.batch = batch;
// https://stackoverflow.com/questions/9539513/is-there-a-reliable-way-in-javascript-to-obtain-the-number-of-decimal-places-of?noredirect=1&lq=1
function decimalPlaces(n) {
    // Make sure it is a number and use the builtin number -> string.
    var s = '' + +n;
    // Pull out the fraction and the exponent.
    var match = /(?:\.(\d+))?(?:[eE]([+\-]?\d+))?$/.exec(s);
    // NaN or Infinity or integer.
    // We arbitrarily decide that Infinity is integral.
    if (!match) {
        return 0;
    }
    // Count the number of digits in the fraction and subtract the
    // exponent to simulate moving the decimal point left by exponent places.
    // 1.234e+2 has 1 fraction digit and '234'.length -  2 == 1
    // 1.234e-2 has 5 fraction digit and '234'.length - -2 == 5
    return Math.max(0, // lower limit.
    (match[1] == '0' ? 0 : (match[1] || '').length) - // fraction length
        (+match[2] || 0)); // exponent
}
exports.decimalPlaces = decimalPlaces;
class CircularBuffer {
    _bufferSize;
    _buffer = [];
    _index = 0;
    constructor(_bufferSize) {
        this._bufferSize = _bufferSize;
    }
    append(value) {
        const isFull = this._buffer.length === this._bufferSize;
        let poppedValue;
        if (isFull) {
            poppedValue = this._buffer[this._index];
        }
        this._buffer[this._index] = value;
        this._index = (this._index + 1) % this._bufferSize;
        return poppedValue;
    }
    *items() {
        for (let i = 0; i < this._buffer.length; i++) {
            const index = (this._index + i) % this._buffer.length;
            yield this._buffer[index];
        }
    }
    get count() {
        return this._buffer.length;
    }
    clear() {
        this._buffer = [];
        this._index = 0;
    }
}
exports.CircularBuffer = CircularBuffer;
const { BroadcastChannel } = require('worker_threads');
exports.minionReadyChannel = new BroadcastChannel('MinionReady');
exports.zetaProducerReadyChannel = new BroadcastChannel('ZetaProducerReady');
exports.zetaDataChannel = new BroadcastChannel('ZetaData');
exports.zetaMarketsChannel = new BroadcastChannel('ZetaMarkets');
exports.cleanupChannel = new BroadcastChannel('Cleanup');
async function executeAndRetry(operation, { maxRetries }) {
    let attempts = 0;
    while (true) {
        attempts++;
        try {
            return await operation(attempts);
        }
        catch (err) {
            if (attempts > maxRetries) {
                throw err;
            }
            await (0, exports.wait)(500 * attempts * attempts);
        }
    }
}
exports.executeAndRetry = executeAndRetry;
async function getZetaPerpMarkets(cluster, endpoint, throttleMs) {
    const programId = cluster === 'mainnet-beta'
        ? new web3_js_1.PublicKey('ZETAxsqBRek56DhiGXrn75yj2NHU3aYUnxvHXpkf3aD') // mainnet progranId
        : new web3_js_1.PublicKey('BG3oRikW8d16YjUEmX3ZxHm9SiJzrGtMhsSR8aCw1Cd7'); // devnet programId
    const network = cluster == 'mainnet-beta' ? sdk_1.Network.MAINNET : sdk_1.Network.DEVNET;
    const defaultUrl = cluster == 'mainnet-beta' ? sdk_1.constants.CLUSTER_URLS.mainnet : sdk_1.constants.CLUSTER_URLS.devnet;
    const networkUrl = !endpoint || endpoint === 'default' ? defaultUrl : endpoint;
    const connection = new web3_js_1.Connection(networkUrl, sdk_1.utils.defaultCommitment());
    await sdk_1.Exchange.load(sdk_1.assets.allAssets(), programId, network, connection, sdk_1.utils.defaultCommitment(), undefined, throttleMs || 1000, undefined);
    const allEx = sdk_1.Exchange.getAllSubExchanges();
    const sortedEx = allEx
        .map((ex) => {
        return ex.markets.markets
            .filter((mkt) => mkt.kind == 'future')
            .map((mrkt) => {
            return {
                name: (0, assets_1.assetToName)(mrkt.asset) + '-' + mrkt.marketIndex,
                alias: [(0, assets_1.assetToName)(mrkt.asset) + '-' + mrkt.marketIndex, (0, assets_1.assetToName)(mrkt.asset) + '/' + mrkt.expiryIndex],
                asset: mrkt.asset,
                expiryIndex: mrkt.expiryIndex,
                marketIndex: mrkt.marketIndex,
                address: mrkt.address.toBase58(),
                dexPId: mrkt.serumMarket.programId.toBase58()
            };
        });
    })
        .flat();
    sdk_1.Exchange.close();
    return sortedEx;
}
exports.getZetaPerpMarkets = getZetaPerpMarkets;
//# sourceMappingURL=helpers.js.map