"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.SerumProducer = void 0;
const serum_1 = require("@project-serum/serum");
const web3_js_1 = require("@solana/web3.js");
const worker_threads_1 = require("worker_threads");
const data_mapper_1 = require("./data_mapper");
const helpers_1 = require("./helpers");
const logger_1 = require("./logger");
const rpc_client_1 = require("./rpc_client");
if (worker_threads_1.isMainThread) {
    const message = 'Exiting. Worker is not meant to run in main thread';
    logger_1.logger.log('error', message);
    throw new Error(message);
}
process.on('unhandledRejection', (err) => {
    throw err;
});
// SerumProducer responsibility is to:
// - connect to Serum Node RPC API via WS and subscribe to single Serum market
// - map received data to normalized data messages and broadcast those
class SerumProducer {
    _options;
    constructor(_options) {
        this._options = _options;
    }
    async run(onData) {
        let started = false;
        logger_1.logger.log('info', `Serum producer starting for ${this._options.market.name} market...`);
        // don't use Solana web3.js Connection but custom rpcClient so we have more control and insight what is going on
        const rpcClient = new rpc_client_1.RPCClient({
            nodeEndpoint: this._options.nodeEndpoint,
            commitment: this._options.commitment,
            wsEndpointPort: this._options.wsEndpointPort
        });
        const market = await serum_1.Market.load(rpcClient, new web3_js_1.PublicKey(this._options.market.address), undefined, new web3_js_1.PublicKey(this._options.market.dexPId));
        const priceDecimalPlaces = (0, helpers_1.decimalPlaces)(market.tickSize);
        const sizeDecimalPlaces = (0, helpers_1.decimalPlaces)(market.minOrderSize);
        const dataMapper = new data_mapper_1.DataMapper({
            symbol: this._options.market.name,
            market,
            priceDecimalPlaces,
            sizeDecimalPlaces
        });
        let start = process.hrtime();
        const interval = 600;
        // based on https://github.com/tj/node-blocked/blob/master/index.js
        setInterval(() => {
            const delta = process.hrtime(start);
            const nanosec = delta[0] * 1e9 + delta[1];
            const ms = nanosec / 1e6;
            const n = ms - interval;
            if (n > 200) {
                logger_1.logger.log('info', `Event loop blocked for ${Math.round(n)} ms.`, {
                    market: this._options.market.name
                });
            }
            start = process.hrtime();
        }, interval).unref();
        for await (const notification of rpcClient.streamAccountsNotification(market, this._options.market.name)) {
            if (started === false) {
                logger_1.logger.log('info', `Serum producer started for ${this._options.market.name} market...`);
                started = true;
                helpers_1.serumProducerReadyChannel.postMessage('ready');
            }
            const messagesForSlot = [...dataMapper.map(notification)];
            if (messagesForSlot.length > 0) {
                onData(messagesForSlot);
            }
        }
    }
}
exports.SerumProducer = SerumProducer;
const serumProducer = new SerumProducer(worker_threads_1.workerData);
serumProducer.run((envelopes) => {
    helpers_1.serumDataChannel.postMessage(envelopes);
});
//# sourceMappingURL=serum_producer.js.map