import { Connection, PublicKey } from '@solana/web3.js';
import { assets, constants, Exchange, Network, utils } from '@zetamarkets/sdk';
import { assetToName } from '@zetamarkets/sdk/dist/assets';
import didYouMean from 'didyoumean2';
import { ZetaMarket } from './types';

export const wait = (delayMS: number) => new Promise((resolve) => setTimeout(resolve, delayMS));

export function getDidYouMean(input: string, allowedValues: readonly string[]) {
  let tip = '';

  if (typeof input === 'string') {
    let result = didYouMean(input, allowedValues, {});
    if (result !== null) {
      tip = ` Did you mean '${result}'?`;
    }
  }
  return tip;
}

export function getAllowedValuesText(allowedValues: readonly string[]) {
  return `Allowed values: ${allowedValues.map((val) => `'${val}'`).join(', ')}.`;
}

export function* batch<T>(items: T[], batchSize: number) {
  for (let i = 0; i < items.length; i += batchSize) {
    yield items.slice(i, i + batchSize);
  }
}

// https://stackoverflow.com/questions/9539513/is-there-a-reliable-way-in-javascript-to-obtain-the-number-of-decimal-places-of?noredirect=1&lq=1

export function decimalPlaces(n: number) {
  // Make sure it is a number and use the builtin number -> string.
  var s = '' + +n;
  // Pull out the fraction and the exponent.
  var match = /(?:\.(\d+))?(?:[eE]([+\-]?\d+))?$/.exec(s);
  // NaN or Infinity or integer.
  // We arbitrarily decide that Infinity is integral.
  if (!match) {
    return 0;
  }
  // Count the number of digits in the fraction and subtract the
  // exponent to simulate moving the decimal point left by exponent places.
  // 1.234e+2 has 1 fraction digit and '234'.length -  2 == 1
  // 1.234e-2 has 5 fraction digit and '234'.length - -2 == 5

  return Math.max(
    0, // lower limit.
    (match[1] == '0' ? 0 : (match[1] || '').length) - // fraction length
      (+match[2]! || 0)
  ); // exponent
}

export class CircularBuffer<T> {
  private _buffer: T[] = [];
  private _index: number = 0;
  constructor(private readonly _bufferSize: number) {}

  append(value: T) {
    const isFull = this._buffer.length === this._bufferSize;
    let poppedValue;
    if (isFull) {
      poppedValue = this._buffer[this._index];
    }
    this._buffer[this._index] = value;
    this._index = (this._index + 1) % this._bufferSize;

    return poppedValue;
  }

  *items() {
    for (let i = 0; i < this._buffer.length; i++) {
      const index = (this._index + i) % this._buffer.length;
      yield this._buffer[index]!;
    }
  }

  get count() {
    return this._buffer.length;
  }

  clear() {
    this._buffer = [];
    this._index = 0;
  }
}

const { BroadcastChannel } = require('worker_threads');

export const minionReadyChannel = new BroadcastChannel('MinionReady') as BroadcastChannel;
export const zetaProducerReadyChannel = new BroadcastChannel('ZetaProducerReady') as BroadcastChannel;
export const zetaDataChannel = new BroadcastChannel('ZetaData') as BroadcastChannel;
export const zetaMarketsChannel = new BroadcastChannel('ZetaMarkets') as BroadcastChannel;
export const cleanupChannel = new BroadcastChannel('Cleanup') as BroadcastChannel;

export async function executeAndRetry<T>(
  operation: (attempt: number) => Promise<T>,
  { maxRetries }: { maxRetries: number }
): Promise<T> {
  let attempts = 0;
  while (true) {
    attempts++;
    try {
      return await operation(attempts);
    } catch (err) {
      if (attempts > maxRetries) {
        throw err;
      }

      await wait(500 * attempts * attempts);
    }
  }
}

export async function getZetaPerpMarkets(
  cluster: 'mainnet-beta' | 'devnet',
  endpoint?: string,
  throttleMs?: number
): Promise<ZetaMarket[]> {

  const programId =
    cluster === 'mainnet-beta'
      ? new PublicKey('ZETAxsqBRek56DhiGXrn75yj2NHU3aYUnxvHXpkf3aD') // mainnet progranId
      : new PublicKey('BG3oRikW8d16YjUEmX3ZxHm9SiJzrGtMhsSR8aCw1Cd7'); // devnet programId

  const network = cluster == 'mainnet-beta' ? Network.MAINNET : Network.DEVNET;

  const defaultUrl = cluster == 'mainnet-beta' ? constants.CLUSTER_URLS.mainnet : constants.CLUSTER_URLS.devnet;

  const networkUrl = !endpoint || endpoint === 'default' ? defaultUrl : endpoint;

  const connection = new Connection(networkUrl, utils.defaultCommitment());

  await Exchange.load(
    assets.allAssets(),
    programId,
    network,
    connection,
    utils.defaultCommitment(),
    undefined,
    throttleMs || 1000,
    undefined
  );

  const allEx = Exchange.getAllSubExchanges();

  const sortedEx = allEx
    .map((ex) => {
      return ex.markets.markets
        .filter((mkt) => mkt.kind == 'future')
        .map((mrkt) => {
          return {
            name: assetToName(mrkt.asset) + '-' + mrkt.marketIndex,
            alias: [assetToName(mrkt.asset) + '-' + mrkt.marketIndex, assetToName(mrkt.asset) + '/' + mrkt.expiryIndex],
            asset: mrkt.asset,
            expiryIndex: mrkt.expiryIndex,
            marketIndex: mrkt.marketIndex,
            address: mrkt.address.toBase58(),
            dexPId: mrkt.serumMarket.programId.toBase58()
          };
        });
    })
    .flat();

  Exchange.close();

  return sortedEx;
}
